#include "includes/Screen.hpp"
#include "../includes/types.hpp"
#include "../includes/asm.hpp"
//#include <stdargs.h>

namespace BootProcess
{
  namespace Screen
  {
    s8 X = 0;
    s8 Y = 0;
    s8 Attributes = 0x07;

    void ScrollUp(u32 lines)
    {
      u8* tmp;

      for (u8* video = (u8*)0xB8000; video < (u8*)0xB8FA0; video += 2)
      {
        tmp = (u8*)(video + lines * 80 * 2);

        if (tmp < (u8*)0xB8FA0)
        {
          *video = *tmp;
          *(video + 1) = *(tmp + 1);          
        }
        else
        {
          *video = 0;
          *(video + 1) = Attributes;          
        }
      }

      Y -= lines;

      if (Y < 0)
        Y = 0;
    }

    void PrintCharacter(s8 character)
    {
      if (character == 10)      //  CR
        Y++;
      else if (character == 13) //  LF
        X = 0;
      else if (character == 9)  //  TAB
      {
        s8 numberOfSpacesToPrint = 8 - (X % 8);
        for (s8 i = 0; i < numberOfSpacesToPrint; i++)
          PrintCharacter(0x20);

        X += numberOfSpacesToPrint;
      }
      else
      {
        s8* video = (s8*)0xB8000+ Y*80*2 + X*2;
        *video = character;
        *(video + 1) = Attributes;        
        X++;
      }

      if (X > 79)
      {
        X = 0;
        Y++;
      }

      if (Y > 24)
        ScrollUp(Y - 24);
    }

    void PrintCharacterXY(s8 character, s8 x, s8 y)
    {
      s8 currentX = X;
      s8 currentY = Y;

      X = x;
      Y = y;

      PrintCharacter(character);

      X = currentX;
      Y = currentY;
    }

    void Print(s8* string)
    {
      s32 i = 0;

      while(string[i] != 0x00)
        PrintCharacter(string[i++]);
    }

    void PrintXY(s8* string, s8 x, s8 y)
    {
      s32 i = 0;

      while(string[i] != 0x00)
      {
        PrintCharacterXY(string[i++], x, y);
        x++;

        if (string[i] == 0x10)
          y++;
        else if (string[i] == 0x13)
          x = 0;
        else if (string[i] == 0x09)
        {
          s32 numSpaces = 8 - (X % 8);
          for (s32 i = 0; i < numSpaces; i++)
            PrintCharacterXY(0x20, x + i, y);
          x += numSpaces;
        }
      }
    }

    void ClearWithColor(s8 attributes)
    {
      for (u8* video = (u8*)0xB8000; video < (u8*)0xB8FA0; video += 2)
      {
        *video = 0;
        *(video + 1) = attributes;        
      }

      X = 0;
      Y = 0;
    }

    void Clear()
    {
      ClearWithColor(Attributes);
    }

    void MoveCursor(u8 x, u8 y)
    {
      u16 cursorPosition = y * 80 + x;

      ASM::OUTB(0x3D4, 0x0F);
      ASM::OUTB(0x3D5, (u8)(cursorPosition & 0x00FF));
      ASM::OUTB(0x3D4, 0x0E);
      ASM::OUTB(0x3D5, (u8)(cursorPosition >> 8));
    }

    void ShowCursor()
    {
      MoveCursor((u8)X, (u8)Y);
    }

    void HideCursor()
    {
      MoveCursor((u8)-1, (u8)-1);
    }

    void PrintFormat(s8* s, ...)
    {
    	va_list ap;
    
    	s8 buf[16];
    	s8 i, j, size, buflen, neg;
    
    	u8 c;
    	s32 ival;
    	u32 uival;
    
    	va_start(ap, s);
    
    	while ((c = *s++))
      {
    		size = 0;
    		neg = 0;
    
    		if (c == 0)
    			break;
    		else if (c == '%')
        {
    			c = *s++;
    			if (c >= '0' && c <= '9')
          {
    				size = c - '0';
    				c = *s++;
    			}
    
    			if (c == 'd')
          {
    				ival = va_arg(ap, int);
    				if (ival < 0)
            {
    					uival = 0 - ival;
    					neg++;
    				}
            else
    					uival = ival;
    
    				itoa(buf, (unsigned long int)uival, 10);
    
    				buflen = strlen(buf);
    				if (buflen < size)
    					for (i = size, j = buflen; i >= 0; i--, j--)
    						buf[i] = (j >= 0) ? buf[j] : '0';
    
    				if (neg)
    					PrintFormat((s8*)"-%s", buf);
    				else
    					Print(buf);
    			}
          else if (c == 'u')
          {
    				uival = va_arg(ap, int);
    				itoa(buf, (unsigned long int)uival, 10);
    
    				buflen = strlen(buf);
    				if (buflen < size)
    					for (i = size, j = buflen; i >= 0; i--, j--)
    						buf[i] = (j >= 0) ? buf[j] : '0';
    
    				Print(buf);
    			}
          else if (c == 'x' || c == 'X')
          {
    				uival = va_arg(ap, int);
    				itoa(buf, (unsigned long int)uival, 16);
    
    				buflen = strlen(buf);
    				if (buflen < size)
    					for (i = size, j = buflen; i >= 0; i--, j--)
    						buf[i] = (j >= 0) ? buf[j] : '0';
    
    				PrintFormat((s8*)"%s", buf);
    			}
          else if (c == 'p')
          {
    				uival = va_arg(ap, int);
    				itoa(buf, (unsigned long int)uival, 16);
    				size = 8;
    
    				buflen = strlen(buf);
    				if (buflen < size)
    					for (i = size, j = buflen; i >= 0; i--, j--)
    						buf[i] = (j >= 0) ? buf[j] : '0';
    
    				PrintFormat((s8*)"0x%s", buf);
    			}
          else if (c == 's')
          {
    				PrintFormat((s8*) va_arg(ap, int));
    			} 
    		}
        else
    			PrintCharacter(c);
    	}
    
    	return;
    }

    void itoa(s8 *buf, unsigned long int n, int base)
    {
    	unsigned long int tmp;
    	int i, j;
    
    	tmp = n;
    	i = 0;
    
    	do
      {
    		tmp = n % base;
    		buf[i++] = (tmp < 10) ? (tmp + '0') : (tmp + 'a' - 10);
    	} while (n /= base);
    	buf[i--] = 0;
    
    	for (j = 0; j < i; j++, i--)
      {
    		tmp = buf[j];
    		buf[j] = buf[i];
    		buf[i] = tmp;
    	}
    }

    int strlen(s8* s)
    {
    	int i = 0;
    
    	while (*s++)
    		i++;
    
    	return i;
    }

  }
}
