#include "includes/PhysicalMemoryManager.hpp"
#include "includes/types.hpp"
#include "includes/MultiBoot.hpp"

namespace PhysicalMemoryManager
{
  //  Every bit represent a page frame allocation of 4096 bytes (4 ko pages frame)
  //    Total bitmap memory manageable with this: 4 Go (32 * 4096 * 32768)
  //    Total memory for this: 32768 * 4 == 131072 bytes (128 ko)
  u32 Bitmap[32768];

  void Init(struct MultiBoot::Informations* multiBootInformations)
  {
    // First, initialize all memory to "used" state (1)
    for (s32 i = 0; i < 32768; i++)
      Bitmap[i] = 0xFFFFFFFF;

    // Second, initialize memory marked "available" (from GRUB) with the "free state" (0)
    struct MultiBoot::InformationsMemoryMap* memoryMap = (struct MultiBoot::InformationsMemoryMap*)multiBootInformations->MemoryMapAddress;
    while ((u32)memoryMap < multiBootInformations->MemoryMapAddress + multiBootInformations->MemoryMapLength)
    {
      if (memoryMap->Type == 0x00000001)  //  "Available" memory
      {
        // Delete last bytes if not multiple of 4096 bytes. They will are not used.
        u32 length = memoryMap->Length / 4096;
        length *= 4096;

        u32 currentAddress = memoryMap->BaseAddress;

        // We want compute only addresses which are multiple of 4096
        if (currentAddress % 4096 == 0)
        {
          currentAddress /= 4096;
          currentAddress = (currentAddress * 4096) + 4096;
        }

        u32 startAddress = currentAddress;

        while(currentAddress < (startAddress + length))
        {
          SetPageFrameAvailable(currentAddress);
          currentAddress += 4096;
        }
      }

      memoryMap = (struct MultiBoot::InformationsMemoryMap*)((u32)memoryMap + memoryMap->Size + sizeof(u32));
    }

    // Third, mark address 0x00000000 as "used" for return NULL (0x00) with GetFirstPageFrameAvailable()
    //  There is also here the IDT (2048 o) and the GDT (2048 o)
    SetPageFrameUsed(0x00000000);

    // Fourth, mark "used" the kernel base memory (we start @ 0x100000 = 1 Mo)
    int SizeOfKernel = 4096*1024;   //  Assume kernel size is 4 Mb
    for (int i = 0; i < SizeOfKernel; i += 4096)
      SetPageFrameUsed(0x100000 + (i));   //  Base Address of kernel (4k aligned) + 4k block(s)
  }

  void SetPageFrameUsed(u32 physicalAddress)
  {
    u32 bitNumber = physicalAddress / 4096;
    Bitmap[bitNumber / 32] |= (0x01 << (bitNumber % 32));
  }

  void SetPageFrameAvailable(u32 physicalAddress)
  {
    u32 bitNumber = physicalAddress / 4096;
    Bitmap[bitNumber / 32] &= (~(0x01 << (bitNumber % 32)));
  }

  u32 GetFirstPageFrameAvailable()
  {
    for (u32 i = 0; i < 32768; i++)
      if (Bitmap[i] != 0xFFFFFFFF)
        for (u32 j = 0; j < 32; j++)
          if ((Bitmap[i] & (0x01 << j)) == 0x00)
            return (i * 32 * 4096 + j * 4096);

    return 0x00000000;
  }
}
