#include "includes/VirtualMemoryManager.hpp"
#include "includes/types.hpp"
#include "includes/PhysicalMemoryManager.hpp"

#if DEBUG
  #include "BootProcess/includes/Screen.hpp"
#endif

namespace VirtualMemoryManager
{
  struct PageDirectory* KernelPageDirectory = 0x00;
  struct PageDirectory* CurrentPageDirectory  = 0x00;

  void Init()
  {
    //  Kernel Page Directory creation
    KernelPageDirectory = NewPageDirectory(0);  //  Assume that we have place in memory

    // Page Tables allocation, all memory mapped as identity (0 => 0, 1 Go = 1 Go)
    for (int i = 0; i < 1024; i++)
    {
      struct PageTable* kernelPageTable = NewPageTable(0);  //  Assume that we have place in memory

      for (int j = 0; j < 1024; j++)
      {
        kernelPageTable->Entries[j].Present = 1;
        kernelPageTable->Entries[j].UserSupervisor = 0;
        kernelPageTable->Entries[j].ReadWrite = 1;
        //kernelPageTable->Entries[j].PAT = 0;
        //kernelPageTable->Entries[j].Dirty = 0;
        kernelPageTable->Entries[j].PhysicalAddress = ((4096 * 1024 * i) + (4096 * j)) >> 12;
      }

      KernelPageDirectory->Entries[i].Present = 1;
      KernelPageDirectory->Entries[i].UserSupervisor = 0;
      KernelPageDirectory->Entries[i].ReadWrite = 1;
      //KernelPageDirectory->Entries[i].PageSize = 0;
//      KernelPageDirectory->Entries[i].PhysicalAddress = ((u32)(kernelPageTable)) >> 12;
      u32 address = ((u32)&kernelPageTable[0]) >> 12;
      KernelPageDirectory->Entries[i].PhysicalAddress = address;
    }

    /*#if DEBUG
      BootProcess::Screen::Print((s8*)"\r\n******* Ok HERE ********\r\n");
      for(int i = 0; i < 100000000; i++);
    #endif*/

    //  Enable Paging
    asm("mov %0, %%eax          \n  \
         mov %%eax, %%cr3       \n  \
         mov %%cr0, %%eax       \n  \
         or $0x80000000, %%eax  \n  \
         mov %%eax, %%cr0" : : "m"(KernelPageDirectory));


    //  DEBUG: It's working great!
/*#define MAX 256

    struct PageDirectory* pd[MAX];
    for (int i = 0; i < 32; i++)
    {
      for (int j = 0; j < MAX; j++)
      {
          pd[j] = NewPageDirectory(0);
          printk("DEBUG: %p\n", pd[j]);
      }
      for (int j = 0; j < MAX; j++)
        ReleasePageDirectory(pd[j]);
    }*/                             

    CurrentPageDirectory = KernelPageDirectory;
  }

  struct PageDirectory* NewPageDirectory(u8 user)
  {
    struct PageDirectory* newPageDirectory = (struct PageDirectory*)PhysicalMemoryManager::GetFirstPageFrameAvailable();

    if (newPageDirectory == 0x00)
      return newPageDirectory;

    PhysicalMemoryManager::SetPageFrameUsed((u32)newPageDirectory);

    for (int i = 0; i < 1024; i++)
    {
      newPageDirectory->Entries[i].Present = 0;
      newPageDirectory->Entries[i].ReadWrite = 0;
      newPageDirectory->Entries[i].UserSupervisor = user;
      /*newPageDirectory->Entries[i].PageLevelWriteThrough = 0;
      newPageDirectory->Entries[i].PageLevelCacheDisable = 0;
      newPageDirectory->Entries[i].Accessed = 0;
      newPageDirectory->Entries[i].Ignored = 0;
      newPageDirectory->Entries[i].PageSize = 0;
      newPageDirectory->Entries[i].Ignored2 = 0;*/
      newPageDirectory->Entries[i].PhysicalAddress = 0x00000;
    }

    return newPageDirectory;
  }

  void ReleasePageDirectory(struct PageDirectory* pageDirectory)
  {
    for (int i = 0; i < 1024; i++)
    {
      pageDirectory->Entries[i].Present = 0;
      pageDirectory->Entries[i].PhysicalAddress = 0x00000;
    }

    PhysicalMemoryManager::SetPageFrameAvailable((u32)pageDirectory);
    pageDirectory = 0x00;
  }

  struct PageTable* NewPageTable(u8 user)
  {
    struct PageTable* newPageTable = (struct PageTable*)PhysicalMemoryManager::GetFirstPageFrameAvailable();

    if (newPageTable == 0x00)
      return newPageTable;

    PhysicalMemoryManager::SetPageFrameUsed((u32)newPageTable);

    for (int i = 0; i < 1024; i++)
    {
      newPageTable->Entries[i].Present = 0;
      newPageTable->Entries[i].ReadWrite = 0;
      newPageTable->Entries[i].UserSupervisor = user;
      /*newPageTable->Entries[i].PageLevelWriteThrough = 0;
      newPageTable->Entries[i].PageLevelCacheDisable = 0;
      newPageTable->Entries[i].Accessed = 0;
      newPageTable->Entries[i].Dirty = 0;
      newPageTable->Entries[i].PAT = 0;
      newPageTable->Entries[i].Global = 0;
      newPageTable->Entries[i].Ignored = 0;*/
      newPageTable->Entries[i].PhysicalAddress = 0x00000;
    }

    return newPageTable;
  }

  void ReleasePageTable(struct PageTable* pageTable)
  {
    for (int i = 0; i < 1024; i++)
    {
      pageTable->Entries[i].Present = 0;
      pageTable->Entries[i].PhysicalAddress = 0x00000;
    }

    PhysicalMemoryManager::SetPageFrameAvailable((u32)pageTable);
    pageTable = 0x00;
  }

  void AddPageTable(struct PageDirectory* pageDirectory, struct PageTable* pageTable, u32 virtualAddress)
  {
    u32 pageDirectoryIndex = virtualAddress >> 22;

    pageDirectory->Entries[pageDirectoryIndex].PhysicalAddress = ((u32)pageTable >> 12);
    pageDirectory->Entries[pageDirectoryIndex].Present = 1;                       
  }

  void InitPageTable(struct PageTable* pageTable, int user)
  {
    for (int i = 0; i < 1024; i++)
    {
      pageTable->Entries[i].UserSupervisor = user;
      pageTable->Entries[i].PhysicalAddress = 0x00000;
      pageTable->Entries[i].Present = 0;
    }
  }

  void InitUserPageDirectory(PageDirectory* pageDirectory)
  {
    for(int i = 0; i < 256; i++)
      pageDirectory->Entries[i] = KernelPageDirectory->Entries[i];

    for(int i = 256; i < 1024; i++)
    {
      pageDirectory->Entries[i].Present = 0;
      pageDirectory->Entries[i].PhysicalAddress = 0x00000;
      pageDirectory->Entries[i].UserSupervisor = 1;
    }
  }

  void AddPhysicalAddress(PageTable* pageTable, u32 virtualAddress, u32 physicalAddress)
  {
    u32 pageTableIndex = (virtualAddress >> 12) & 0x03FF;

    pageTable->Entries[pageTableIndex].PhysicalAddress = physicalAddress >> 12;
    pageTable->Entries[pageTableIndex].Present = 1;
  }
}
